 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */
package Java2OWL;


import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.logging.ConsoleHandler;
import java.util.logging.FileHandler;
import java.util.logging.Filter;
import java.util.logging.Formatter;
import java.util.logging.Handler;
import java.util.logging.Level;
import java.util.logging.LogRecord;
import java.util.logging.Logger;
import java.util.logging.StreamHandler;

/** This is an interface to Java's Logger system.
 *  <p>Each J2OManager gets its own logger, which can be individually configured.
 *  Log entries are submitted to the logger by one of the methods info, error etc.
 *  They are semantically classified and can be selectively turned on and off.
 * 
 * @author ohlbach
 */
public class Logging { 
    /** the top-level manager */
    private J2OManager manager;     
        
    /** the index of the top-level manager (for distinguishing the messages from the different managers or threads */
    private int counter;            
    
    /** Java's logger */
    private Logger logger;          
  
    /** activates/deactivates all loggings which are not errors or warnings. */
    private boolean infoActivated;   
    
    /** for in-memory logging */
    private ByteArrayOutputStream byteArray = null;
    
    /** the handler for the logger. */
    private Handler handler = null;   
    
    /** the file name if it is a file logger. */
    private String filename = null;
    
    /** indicates whether errors have been logged. */
    private boolean errorsLogged = false;
    
    /** indicates that the handler must be closed at the end (System.out need not be closed). */
    private boolean handlerMustClose = false; 
    
    /** indicates debug mode. In debug mode, error messages cause the system to terminate. */
    private boolean debugMode = false;
    
    /** the name of the logger */
    String name = "";
    
    /** puts the logger in/out of debug mode.
     * 
     * @param flag the debug mode.
     * <p> In debug mode the first error signalled to the logger causes the stack trace to be printed
     * and the program to be terminated.
     */
    void setDebugMode(boolean flag) {
        debugMode = flag;}
    
    /** indicates the destination of the output */
    enum Destination {
        /** System.out */
        OUT,
        /** System.err */
        ERR,
        /** in-memory */
        BUFFER,
        /** printed to a file */
        FILE
    }
    
    /** determines the semantics of the messages and can be used to 
     *  activate/deactivate the logging separately for each class of messages. 
     */
    enum Semantics {
        /**information messages */
        INFO("Info"),        
        /** entering, exiting, installation messages */
        CONTROLFLOW("Controlflow"),   
        /** OWL class creation*/
        OCC("OWL-Class creation"), 
        /** OWL property creation*/
        OPC("OWL-Property creation"),
        /** OWL individual creation*/
        OIC("OWL-Indiviudual creation"),
        /** Java object creation*/
        JOC("Java-Object creation"),
        /** data change*/
        DC("Data Change"),
        /** error and warning messages*/
        ERROR("Error");         
        
        private String name;
        public String getName() {return name;}
        Semantics(String name) {this.name = name;}
    }
    
    /** flags for indicating activation/deactivation of the different messages */
    private Map<Semantics,Boolean> activation = new HashMap<Semantics,Boolean>();
    
    /** specifies the origin of error messages. */
    enum Origin {
        USAGE,
        JAVA,
        ONTOLOGY,
        INDIVIDUAL,
        REASONER,
        UNCLEAR;
    }
    /** Creates a new logger for the J2OManager 
     * 
     * @param manager the top-level manager
     * <p>The default handler is a stream handler which sends all messages to System.out
     * All message classes are activated.
     */
     
    public Logging(J2OManager manager) {
        if(manager == null) {
            this.manager = null;
            debugMode = false;
            counter = 0;
            name = "Logger";}
        else {
            this.manager = manager;
            debugMode = manager.isInDebugMode(); 
            counter = manager.getNumber();
            name = manager.getName() + "_Logger";}
        
        
        logger = Logger.getLogger(name);
        logger.setUseParentHandlers(false);
        logger.setLevel(Level.ALL);
        handler = new StreamHandler(System.out,new J2OFormatter());
        handler.setFilter(filter);
        logger.addHandler(this.handler);
        infoActivated = true;
        activate(true);
        this.installation(this); // logs its own installation.
    }
    
    /** returns the name of the logger. */
    public String getName() {return name;}
    
    /** allows one to change the handler.
     * 
     * @param where either OUT, ERR of BUFFER or FILE.
     * @param filename is a filename, if necessary.
     * @throws IOException if the file could not be opened.
     * <p>The handler sends the messages either to System,out, System.err, the given file or to the internal buffer.
     */
    public synchronized void setHandler(Logging.Destination where, String filename) throws IOException {
        this.filename = null;
        if(handlerMustClose) handler.close();
        handlerMustClose = false;
        errorsLogged = false;
        byteArray = null;
        logger.removeHandler(handler);
        Formatter formatter = handler.getFormatter();
        switch(where) {
            case OUT:  handler = new StreamHandler(System.out,formatter);break;
            case ERR:  handler = new ConsoleHandler(); break;
            case BUFFER: 
                    byteArray = new ByteArrayOutputStream(); 
                    handler = new StreamHandler(byteArray,formatter);
                    break;
            case FILE:  handler = new FileHandler(filename); 
                        this.filename = filename;
                        handlerMustClose = true;}
        handler.setFormatter(formatter);
        handler.setFilter(filter);
        logger.addHandler(handler);
    }
    
    /** allows one to install a user defined handler.
     * 
     * @param handler the custom handler
     * @param closable true if the handler must be closed at the end
     * <p>The old handler is removed.
     * The new handler gets the formatter of the previous handler.
     */
    public synchronized void setHandler(Handler newHandler,boolean closable) {
        filename = null;
        if(handlerMustClose) handler.close();
        handlerMustClose = closable;
        errorsLogged = false;
        byteArray = null;
        logger.removeHandler(handler);
        newHandler.setFormatter(handler.getFormatter());
        handler = newHandler;
        newHandler.setFilter(filter);
        logger.addHandler(newHandler);}
    
   
    /** returns the loggings as string if the handler was set as "BUFFER". */
    public synchronized String getContents() {
        handler.flush();
        return (byteArray == null)  ? "" : byteArray.toString();}
    
    /** returns true if the buffer is not empty .*/
    public synchronized boolean hasContents() {
        handler.flush();
        return byteArray != null && !byteArray.toString().isEmpty();}
    
    /** clears the internal buffer or closes and reopens the file. */ 
    public synchronized void clear() {
        handler.flush();
        if(handlerMustClose) handler.close();
        errorsLogged = false;
        if(byteArray != null) byteArray.reset();
        if(filename != null) {
            try {setHandler(Logging.Destination.FILE, filename);} 
            catch (IOException ex) {ex.printStackTrace();}}}
    
    /** returns true if errors have been logged. */
    public synchronized boolean hasErrors() {return errorsLogged;}
    
    /** allows one to install a user defined formatter. */
    public synchronized void setFormatter(Formatter formatter) {handler.setFormatter(formatter);}
    
    /** resets the formatter. */
    public synchronized void resetFormatter() {handler.setFormatter(new J2OFormatter());}
     
    /** checks whether logging for the messages with the given semantics is active. */ 
    public boolean isActive(Semantics semantics) {
        return activation.get(semantics);} 
    
    /** activates logging only for messages with the given semantics, and errors and warnings.*/
    public synchronized void activate(Semantics... semantics) {
        for(Semantics sem : Semantics.values()) activation.put(sem,false);
        for(Semantics sem: semantics)activation.put(sem,true);
        activation.put(Semantics.ERROR,true);}
    
    /** activates/deactivates logging for all messages.*/
    public synchronized final void activate(boolean active) {
        for(Semantics semantics : Semantics.values()) activation.put(semantics,active);}
     
    /** globally activates/deactivates all info-messages, not the error-messages. */
    public void activateInfo(boolean active) {
        infoActivated = active;}
   
             
    /** returns the semantics stored in the record. */
    private Semantics getSemantics(LogRecord record) {return (Semantics)(record.getParameters()[0]);}
             
    /** creates a filter which filters messages according to the activated semantics. */
    private Filter filter = new Filter() {
                @Override
                public boolean isLoggable(LogRecord record) {
                    return isActive(getSemantics(record));}};
    
    /** specifies a specific formatter. */
    private class J2OFormatter extends Formatter {
        
      /** formats an array of objects */
      private String formatObjects(Object[] objects) {
        String s = "";
        if(objects.length == 0) return s;
        Throwable exception = null;
        for(Object object : objects) {
            if(object instanceof Throwable) exception = (Throwable)object;
            else s += ((object == null) ? "null" : object.toString()) + ", ";
        }
        if(exception != null) {
            s += "          \n";
            if(exception.getMessage() != null) s += exception.getMessage() + "\n";
            s += Utilities.stackTraceString(exception);}
        else s = s.substring(0,s.length()-2);
        return s;}
            
        /** formats entering/exiting messages. */
        private String formatTrace(LogRecord record) {
            StackTraceElement frame = Utilities.findStackFrame("Java2OWL.Logging");
            String method = "";
            if(frame != null) {
                String classname = frame.getClassName();
                classname = classname.substring(classname.lastIndexOf(".")+1,classname.length());
                method= classname + "." + frame.getMethodName();}
            String message = record.getMessage();
            String date = new Date(record.getMillis()).toString();
            String parameters = formatObjects((Object[])((record.getParameters())[1]));
            return message +  " " +method  + " with " +parameters + " (" + date + ")\n"; 
        }
        
        /** formats installation messages. */
         private String formatInstallation(LogRecord record) {
             String message = record.getMessage();
             Object system = record.getParameters()[1];
             String name = Utilities.getName(system);
             String date = new Date(record.getMillis()).toString();
             return system.getClass().getSimpleName() + " " +name + " " + message + " (" + date + ")\n";}
        
        /** formats info messages. */
        private String formatInfo(LogRecord record) {
             Semantics semantics = (Semantics)(record.getParameters()[0]);
             String date = new Date(record.getMillis()).toString();
             String s = semantics.getName() + ": " +formatObjects((Object[])(record.getParameters()[1]));
             return s + " (" + date + ")\n";}
        
        /** formats error and warning messages. */
        private String formatError(LogRecord record) {
             StackTraceElement frame = Utilities.findStackFrame("Java2OWL.Logging");
             String method = "";
             if(frame != null) method = ", in Method " +frame.getClassName() + "." + frame.getMethodName();
             String message = (String)(record.getParameters()[1]);
             Origin origin = (Origin)(record.getParameters()[2]);
             String date = new Date(record.getMillis()).toString();
             String s = record.getMessage() + " " + message + ", origin " + origin.name() + method +": ";
             Object[] objects = (Object[])(record.getParameters()[3]);
             s += formatObjects(objects); 
             return s + " (" + date + ")\n";}
        
        /** formats correlation messages. */
        private String formatCorrelation(LogRecord record) {
             Semantics semantics = (Semantics)(record.getParameters()[0]);
             String date = new Date(record.getMillis()).toString();
             String s = record.getMessage()+ ": " + semantics.getName() + ": ";
             s += formatObjects((Object[])(record.getParameters()[1]));
             return s + "  (" + date + ")\n";}
        
         
        /** formats a message. */
        @Override
        public String format(LogRecord record) { 
            String prefix = "J2OLOG_" + counter + ": ";
            String message = record.getMessage();
            if(message == ENTERING || message == EXITING) return prefix + formatTrace(record);
            if(message == INSTALLATION) return prefix + formatInstallation(record);
            if(message == INFO) return prefix + formatInfo(record);
            if(message == ERROR || message == WARNING) return prefix + formatError(record);
            if(message == CORRELATION) return prefix + formatCorrelation(record);
            return prefix + formatMessage(record);
            
        }
      }
    
    /** flushes the messages to the output stream. */
    public void flush() {handler.flush();}
    
    static final String ENTERING     = "Entering";
    static final String EXITING      = "Exiting ";
    static final String INSTALLATION = "installed ";
    static final String INFO         = "info";
    static final String ERROR        = "ERROR";
    static final String WARNING      = "WARNING";
    static final String CORRELATION  = "Correlation";
     
    /** indicates entering a method and prints the objects. */
    public synchronized void entering(Object... objects) {
        if(infoActivated) {
            logger.log(Level.INFO,ENTERING,new Object[]{Semantics.CONTROLFLOW,objects});
            if(debugMode) {handler.flush();}}}
    
    /** indicates exiting a method and prints the objects. */
    public synchronized void exiting(Object... objects) {
        if(infoActivated) {
            logger.log(Level.INFO,EXITING,new Object[]{Semantics.CONTROLFLOW,objects});
            if(debugMode) {handler.flush();}}}
    
    /** indicates installation of a module and prints the object. */
    public synchronized void installation(Object object) {
        if(infoActivated) {
            logger.log(Level.INFO,INSTALLATION,new Object[]{Semantics.CONTROLFLOW,object});
            if(debugMode) {handler.flush();}}}
    
     /** any kind of unspecific info message. */
    public synchronized void info(Object... objects) {
        if(infoActivated) {
            logger.log(Level.INFO,INFO,new Object[]{Semantics.INFO,objects});
            if(debugMode) {handler.flush();}}}
     
    /** an error message. In debugMode this method causes the system to stop. */
    public synchronized void error(String message, Origin origin,Object... objects) {
        errorsLogged = true;
        logger.log(Level.SEVERE,ERROR,new Object[]{Semantics.ERROR,message,origin,objects});
        handler.flush();
        if(debugMode) {
            String trace = Utilities.stackTraceString(new Throwable());
            logger.log(Level.SEVERE,ERROR,new Object[]{Semantics.ERROR,"System terminates in debug mode",origin,new Object[]{trace}});
            finalize(); 
            System.exit(1);
        }}
    
    /** an event message with a specific semantics. */
    public synchronized void event(Semantics semantics, Object... objects) {
        if(infoActivated) {
            logger.log(Level.INFO,INFO,new Object[]{semantics,objects});
            if(debugMode) {handler.flush();}}}
     
    /** a warning. */
    public synchronized void warning(String message, Origin origin, Object... objects) {
        logger.log(Level.WARNING,WARNING,new Object[]{Semantics.ERROR,message,origin,objects});
        handler.flush();}
    
    /** a correlation message. */
     public synchronized void correlation(Semantics semantics, Object... objects) {
        if(infoActivated) {
            logger.log(Level.INFO,CORRELATION,new Object[]{semantics,objects});
            if(debugMode) {handler.flush();}}}
    
     /** flushes the messages and closes the handler */
    @Override
    public void finalize() {
        handler.flush();
        if(handlerMustClose) handler.close();}
    
}
